/*
 * $Id$ - dump the contents of the syscall table under Solaris 2.x
 * Release: 1.0c
 *
 * User Datagram Protocol <udp@low-level.net> 25/11/2000
 *
 * Should be 64-bit clean, but doesn't currently grok the thunking that
 * takes place under 64-bit Solaris (there are two syscall tables, one
 * for the 32-bit hosted environment, and one for the 64-bit native
 * operating environment).
 *
 * Shout outs to: jimjones, zsh/lowlevel crew, ADM, and w00w00.
 *
 * Wanted to get this out in a rush, hence had to use GNU hsearch_r()
 * instead of hand rolled hash algorithm.... tsk...
 *
 ****************************************************************************
 *
 * $Log$
 *
 */
 
#include <sys/types.h>
#include <sys/t_lock.h>
#include <sys/mutex.h>
#include <sys/param.h>
#include <sys/map.h>
#include <sys/buf.h>
#include <sys/uio.h>
#include <sys/stream.h>
#include <sys/systm.h>
#include <sys/select.h>
#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <ctype.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <nlist.h>
#include <kvm.h>
#include <search.h>

/* Get this right for ILP32 vs LP64. */
#if defined(_LP64)
#define KVM_PTR_TYPE uintptr_t
#else
#define KVM_PTR_TYPE u_long
#endif

#define MAX_SYSCALLS 512

#ifndef MAX
#define 	  MAX(a, b) (((a) > (b)) ? (a) : (b))
#endif

#ifndef MIN
#define 	  MIN(a, b) (((a) < (b)) ? (a) : (b))
#endif

extern char *strsep(char **str, const char *delim);

static int build_syscall_hashtable(FILE *fsysnum);
static char *lookup_syscall_by_num(const int sysnum);
static int lookup_syscall_by_name(const char *name);

static int parse_ports(char *ports, fd_set *portset);

static const char name_to_sysnum_file[] = "/etc/name_to_sysnum";

static struct hsearch_data name_to_sysnum;
static struct hsearch_data sysnum_to_name;
static char *progname = NULL;

static struct nlist nl[] = {
	{ "sysent" },
	{ 0 }
};

/*--------------------------------------------------------------------------*/

/* reuse bitsets from select() but use them to count calls.
 * [solaris doesn't have bsd's bitstring.h functionality.]
 */

int main(int argc, char *argv[])
{
	extern char	*optarg;
	extern int	optind;
	FILE 		*fsysnum = NULL;
	kvm_t		*kd = NULL;
	fd_set		which;
	int			i, sysnum;
	int			err = 1;
	struct sysent	*se;
	struct sysent	se_cpy;

	progname = argv[0];
	FD_ZERO(&which);

	/* build a hash table for resolving syscall names and vice versa. */
	hcreate_r(MAX_SYSCALLS, &name_to_sysnum);
	hcreate_r(MAX_SYSCALLS, &sysnum_to_name);
	if ((fsysnum = fopen(name_to_sysnum_file, "r")) == NULL) {
		perror("fopen");
		goto catch_errors;
	}	

	if (build_syscall_hashtable(fsysnum))
		goto catch_errors;

#if 0
	/* placeholder for option functionality: int c */
	while ((c = getopt(argc, argv, "h:")) != EOF) {
		switch (c) {
			case 'h':
			case '?':
				err++;
				break;
			default:
		}
	}
#else
	optind = 1;	
#endif

	if ((argc != 2) || err > 1) {
		(void)fprintf(stderr, "listsyscalls - dump Solaris sysent table to stdout\n");
		(void)fprintf(stderr, "usage: %s somecallbyname,somecallbynumber,callrangelo-callrangehi,...,calln\n",
			progname);
		goto catch_errors;
	}
	err = 1;

	if (parse_ports(argv[optind], &which) < 0)
		goto catch_errors;

	if ((kd = kvm_open(NULL, NULL, NULL, O_RDONLY, argv[0])) == NULL) {
		perror("kvm_open");
		goto catch_errors;
	}

	if (kvm_nlist(kd, &nl[0]) < 0) {
		perror("kvm_nlist");
		goto catch_errors;
	}

	if ((void *)nl[0].n_value == NULL) {
		errno = ENOENT;
		perror("kvm_nlist");
		goto catch_errors;
	}
	se = (struct sysent *) nl[0].n_value;

	/* for each syscall in bitset... */
	/* ffs() is cute, but it doesn't work for >32 bits */
	for (i = 0; i < MAX_SYSCALLS; i++) {
		if (!FD_ISSET(i, &which)) continue;
		sysnum = i;
		se += sysnum;
		
		if (kvm_kread(kd, (KVM_PTR_TYPE) se, (void *)&se_cpy, sizeof(struct sysent)) < 0) {
			perror("kvm_kread");
			goto catch_errors;
		}

		fprintf(stdout, "syscall: %s [%d]  sy_narg: %d sy_flags: 0x%02x\n" \
		"sy_call: %p sy_lock: %p sy_callc: %p\n",
			lookup_syscall_by_num(sysnum), sysnum, (int) se_cpy.sy_narg, (int) se_cpy.sy_flags,
			(void *) se_cpy.sy_call, (void *) se_cpy.sy_lock, (void *) se_cpy.sy_callc
		);
	}

	err = 0;
catch_errors:
	if (fsysnum) fclose(fsysnum);
	if (kd)	kvm_close(kd);
	hdestroy_r(&name_to_sysnum);
	hdestroy_r(&sysnum_to_name);

	if (err) fprintf(stderr, "%s: error occurred.\n", progname);

	exit( (err != 0) ? EXIT_FAILURE : EXIT_SUCCESS );	
}

/*--------------------------------------------------------------------------*/

static int build_syscall_hashtable(FILE *fsysnum)
{
	int		callnum = 0, r = -1, line = 0;
	char	callname[32];
	char	callnumstr[32];
	char	*callname_cpy;
	char	*callnumstr_cpy;
	char	*s;
	ENTRY	item, *oitem;

	/* solaris fscanf() heeds field length according to visual inspection.
	 * format bugs in my code? I think not.
	 */		
	rewind(fsysnum);
	while (!feof(fsysnum)) {
		++line;
		if ((r = fscanf(fsysnum, "%31s\t%31s\n", &callname[0], &callnumstr[0])) == 2) {
			callnum = strtoul(callnumstr, &s, 10);
			r = (s == NULL) ? -1 : 0;
		}
		/* ick, ick, ick */
		item.key = callname_cpy = strdup(callname);
		item.data = (void *) callnum;
		hsearch_r(item, ENTER, &oitem, &name_to_sysnum);
		item.key = callnumstr_cpy = strdup(callnumstr);
		item.data = (void *) callname_cpy;
		hsearch_r(item, ENTER, &oitem, &sysnum_to_name);
		if (r) {
			fprintf(stderr, "%s: error parsing %s at line %d\n", progname,
				name_to_sysnum_file, line);
			break;
		}		
	}
	return r;
}

/*--------------------------------------------------------------------------*/

static char *lookup_syscall_by_num(const int sysnum)
{
	ENTRY	item, *oitem;
	char	key[32];

	snprintf(key, sizeof(key), "%d\0", sysnum);
	item.key = key;
	item.data = (void *)0;
	hsearch_r(item, FIND, &oitem, &sysnum_to_name);
	return(oitem ? oitem->data : NULL);
}

/*--------------------------------------------------------------------------*/

static int lookup_syscall_by_name(const char *name)
{
	ENTRY	item, *oitem;

	item.key = (char *)name;
	item.data = (void *)0;
	hsearch_r(item, FIND, &oitem, &name_to_sysnum);
	return(oitem ? (int)oitem->data : (int)0);
}

/*--------------------------------------------------------------------------*/

/* parse at optind in format 1,2,a-b,c-d,...
 * states: in range, got number, got comma, got delimiter (\0)
 * I prefer mine to nmap's.
 */
static int parse_ports(char *ports, fd_set *portset)
{
	char *pargp;
	char *pnext;
	char *cp, *rp, *s;
	int val1, val2, i, start, end, r = -1;

	pnext = pargp = ports;
	while ((pargp = strsep(&pnext, ", \t\n")) != NULL) {
		if (*pargp != '\0') {				
			/* got 'thing', is it a range? */
			if ((rp = index(pargp, '-'))) {
				/* yes it is.. handle tuple accordingly */
				cp = pargp;
				(void) strsep(&cp, "-");
				if (isdigit((int) *pargp)) {
					val1 = strtoul(pargp, &s, 10);
					if (s == NULL) {
						errno = EINVAL;
						perror("strtoul");
						goto catch_errors;		
					}
				} else {
					if (!(val1 = lookup_syscall_by_name(pargp))) {
						errno = EINVAL;
						goto catch_errors;		
					}
				}

				if (isdigit((int) *cp)) {						
					val2 = strtoul(cp, &s, 10);
					if (s == NULL) {
						errno = EINVAL;
						perror("strtoul");
						goto catch_errors;		
					}						
				} else {
					if (!(val2 = lookup_syscall_by_name(pargp))) {
						errno = EINVAL;
						goto catch_errors;		
					}
				}

				start = MIN(val1, val2);
				end = MAX(val1, val2);

				for (i = start; i <= end; i++) {
					FD_SET(i, portset);
				}

			} else {							
				/* a single digit or syscall name */
				if (isdigit((int) *pargp)) {
					start = strtoul(pargp, &s, 10);
					if (s == NULL) {
						errno = EINVAL;
						perror("strtoul");
						goto catch_errors;		
					}
				} else {
					if (!(start = lookup_syscall_by_name(pargp))) {
						errno = EINVAL;
						goto catch_errors;		
					}
				}

				FD_SET(start, portset);
			}
		}
		pargp = pnext;
	}		
	r = 0;
catch_errors:
	return r;
}

/*--------------------------------------------------------------------------*/
